from __future__ import annotations

import sys
from typing import TYPE_CHECKING, ClassVar, Collection

# as ... to mark for re-export
from httpx import HTTPStatusError as HTTPStatusError

if TYPE_CHECKING:

    class _HTTPStatusError(HTTPStatusError):
        _codes: ClassVar[Collection[int]]

else:
    _HTTPStatusError = HTTPStatusError


__all__ = [
    "AgentNotSetError",
    "HTTPStatusError",  # re-export
    "ClientError",
    "BadRequest",
    "Forbidden",
    "NotFound",
    "Conflict",
    "Teapot",
    "TooManyRequests",
    "ServerError",
]


class AgentNotSetError(RuntimeError):
    pass


def narrow(original: HTTPStatusError) -> HTTPStatusError:
    code = original.response.status_code
    match = (sys.maxsize, HTTPStatusError)
    for exc in globals().values():
        if not issubclass(exc, _HTTPStatusError):
            continue
        if exc is HTTPStatusError:
            continue
        codes = exc._codes  # type: ignore
        if code in codes:
            match = min(match, (len(codes), exc))
    _, best = match
    if best is type(original):
        return original
    # bypass __init__
    _new = best.__new__(best)
    _new.__dict__.update(original.__dict__)
    return _new


class ClientError(_HTTPStatusError):
    """
    :exc:`httpx.HTTPStatusError` for 4XX: Client Error status codes.
    """

    _codes = range(400, 500)


class BadRequest(ClientError):
    """
    :exc:`httpx.HTTPStatusError` for 400: Bad Request status codes.
    """

    _codes = (400,)


class Forbidden(ClientError):
    """
    :exc:`httpx.HTTPStatusError` for 403: Forbidden status codes.
    """

    _codes = (403,)


class NotFound(ClientError):
    """
    :exc:`httpx.HTTPStatusError` for 404: Not Found status codes.
    """

    _codes = (404,)


class Conflict(ClientError):
    """
    :exc:`httpx.HTTPStatusError` for 409: Conflict status codes.
    """

    _codes = (409,)


class Teapot(ClientError):
    """
    :exc:`httpx.HTTPStatusError` for 418: I'm a Teapot status codes.
    """

    _codes = (418,)


class TooManyRequests(ClientError):
    """
    :exc:`httpx.HTTPStatusError` for 429: Too Many Requests status codes.
    """

    _codes = (429,)


class ServerError(_HTTPStatusError):
    """
    :exc:`httpx.HTTPStatusError` for 5XX: Server Error status codes.
    """

    _codes = range(500, 600)



from __future__ import annotations

import asyncio
import threading
from collections import deque
from contextvars import Context, copy_context
from functools import partial
from operator import add
from time import monotonic
from typing import TYPE_CHECKING, Any, Callable, Mapping, TypeVar

if TYPE_CHECKING:
    from typing_extensions import Literal

__all__ = ["ResetLock"]


_KT = TypeVar("_KT")


def _get_as_int(mapping: Mapping[_KT, Any], key: _KT, default: int) -> int:
    try:
        return int(mapping[key])
    except (KeyError, ValueError):
        return default


class _Timer(threading.Timer):
    def when(self) -> float:
        # when the timer completes if started at this very moment
        return monotonic() + self.interval

    def run(self) -> None:
        # freeze .when() at the current time
        self.when = partial(add, monotonic(), self.interval)
        return super().run()


# I would love to make this utilize a semaphore instead
# However, since multiple clients can exist on one machine,
# locking and checking the headers sequentially is a requirement
class ResetLock:
    """
    Combination :class:`asyncio.Lock` and :class:`threading.Lock`
    for ratelimiting purposes.

    Usage::

        lock = ResetLock()
        [async] with lock.maybe_lock(url):
            response = ...
            lock.maybe_defer(response.headers)

    Note that, when using :mod:`httpx`,
    using :class:`RateLimiter` will manage a global lock for you.
    """

    def __init__(self):
        self.__deferred: asyncio.TimerHandle | _Timer | None = None
        self._lock = threading.Lock()
        self._waiters: deque[asyncio.Future[Literal[True]]] = deque()

    def _wake_up_next(self):
        try:
            next_fut = next(iter(self._waiters))
        except StopIteration:
            return
        if next_fut.done():
            return
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            loop = None
        next_loop = next_fut.get_loop()
        if next_loop == loop:
            next_fut.set_result(True)
        else:
            next_loop.call_soon_threadsafe(next_fut.set_result, True)

    async def __aenter__(self) -> None:
        acquire = self._lock.acquire
        future_factory = asyncio.get_running_loop().create_future
        waiters_append = self._waiters.append
        waiters_remove = self._waiters.remove
        while not acquire(False):
            fut = future_factory()
            waiters_append(fut)
            try:
                try:
                    await fut
                finally:
                    waiters_remove(fut)
            except asyncio.CancelledError:
                self._wake_up_next()
                raise

    async def __aexit__(self, *_):
        self.release()

    def __enter__(self) -> bool:
        return self._lock.acquire()

    def __exit__(self, *_):
        self.release()

    @property
    def deferred(self) -> float | None:
        """
        Returns when the lock is scheduled to be released according to internal monotonic clocks,
        or None if the lock isn't deferred.
        """
        if self.__deferred:
            return self.__deferred.when()
        # return None

    # write-only property
    @partial(property, None)
    def _deferred(self, value: asyncio.TimerHandle | _Timer | None):
        assert self.locked()
        if self.__deferred:
            self.__deferred.cancel()
        self.__deferred = value

    @_deferred.deleter
    def _deferred(self) -> None:
        assert self.locked()
        if self.__deferred:
            self.__deferred.cancel()
        self.__deferred = None

    def maybe_defer(self, response_headers: Mapping[str, str]) -> None:
        if not self.locked():
            return
        try:
            call_later = asyncio.get_running_loop().call_later  # type: ignore
        except RuntimeError:

            def call_later(
                delay: float,
                callback: Callable[..., Any],
                *args: Any,
                context: Context | None = None,
            ) -> _Timer | asyncio.TimerHandle:
                if not context:
                    context = copy_context()
                timer = _Timer(delay, context.run, (callback, *args))
                timer.start()
                return timer

        xra = _get_as_int(response_headers, "Retry-After", 0)
        if xra:
            self._deferred = call_later(xra, self._release)
            return
        xrlr = _get_as_int(response_headers, "RateLimit-Remaining", 1)
        if xrlr:
            return
        xrlr = _get_as_int(response_headers, "RateLimit-Reset", 0)
        self._deferred = call_later(xrlr, self._release)

    def locked(self):
        return self._lock.locked()

    def release(self):
        if self.deferred is not None:
            return
        self._wake_up_next()
        self._lock.release()

    def _release(self):
        del self._deferred
        self._wake_up_next()
        self._lock.release()



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing_extensions import Literal

from .lock import ResetLock

__all__ = []
lock = ResetLock()
agent: str = ""


def set_agent(new_agent: str, *, _force: Literal[False] = False) -> None:
    """
    Sets sans' User-Agent header.

    Parameters
    ----------
    agent: str
        The User-Agent header to use.
        Your nation name and a method of contact, like an email address, are recommended.
    _force: bool
        Forcibly override the User-Agent header even after it's already been set.
        Not recommended: https://www.nationstates.net/pages/api.html#terms
    """
    global agent
    if agent and not _force:
        raise RuntimeError("Agent cannot be re-set")
    agent = new_agent



from __future__ import annotations

import sys
from typing import Any, Coroutine, Mapping, overload
from xml.etree import ElementTree as etree

from .client import AsyncClientType, ClientType
from .response import Response
from .url import World

__all__ = ["prepare_and_execute", "indent"]


@overload
def prepare_and_execute(
    client: ClientType, *shards: str | Mapping[str, str], **parameters: str
) -> Response:
    ...


@overload
async def prepare_and_execute(
    client: AsyncClientType, *shards: str | Mapping[str, str], **parameters: str
) -> Response:
    ...


def prepare_and_execute(
    client: ClientType | AsyncClientType,
    *shards: str | Mapping[str, str],
    **parameters: str,
) -> Response | Coroutine[Any, Any, Response]:
    if isinstance(client, AsyncClientType):
        return _prepare_async(client, *shards, **parameters)
    request = World(*shards, **parameters, mode="prepare")
    response = client.get(request)
    token: str = response.xml.find("SUCCESS").text  # type: ignore
    request = World(*shards, **parameters, mode="execute", token=token)
    return client.get(request)


async def _prepare_async(
    client: AsyncClientType, *shards: str | Mapping[str, str], **parameters: str
) -> Response:
    request = World(*shards, **parameters, mode="prepare")
    response = await client.get(request)
    token: str = response.xml.find("SUCCESS").text  # type: ignore
    request = World(*shards, **parameters, mode="execute", token=token)
    return await client.get(request)


if sys.version_info < (3, 9):
    # from: https://github.com/python/cpython/blob/3.11/Lib/xml/etree/ElementTree.py#L1154
    def indent(
        tree: etree.Element | etree.ElementTree, space: str = "  ", level: int = 0
    ):
        if isinstance(tree, etree.ElementTree):
            tree = tree.getroot()
        if level < 0:
            raise ValueError(f"Initial indentation level must be >= 0, got {level}")
        if not len(tree):
            return

        # Reduce the memory consumption by reusing indentation strings.
        indentations = ["\n" + level * space]

        def _indent_children(elem: etree.Element, level: int):
            # Start a new indentation level for the first child.
            child_level = level + 1
            try:
                child_indentation = indentations[child_level]
            except IndexError:
                child_indentation = indentations[level] + space
                indentations.append(child_indentation)

            if not elem.text or not elem.text.strip():
                elem.text = child_indentation

            for child in elem:
                if len(child):
                    _indent_children(child, child_level)
                if not child.tail or not child.tail.strip():
                    child.tail = child_indentation

            # Dedent after the last child by overwriting the previous indentation.
            if not child.tail.strip():  # type: ignore
                child.tail = indentations[level]  # type: ignore

        _indent_children(tree, 0)

else:
    from xml.etree.ElementTree import indent



from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    AsyncContextManager,
    Callable,
    Collection,
    ContextManager,
    Mapping,
    overload,
)

import httpx

if TYPE_CHECKING:
    from typing_extensions import TypedDict

    from httpx._types import (
        AuthTypes,
        CertTypes,
        CookieTypes,
        HeaderTypes,
        ProxiesTypes,
        QueryParamTypes,
        RequestContent,
        RequestData,
        RequestFiles,
        TimeoutTypes,
        URLTypes,
        VerifyTypes,
    )

from .limiter import RateLimiter
from .response import Response

__all__ = [
    "Client",
    "AsyncClient",
    "delete",
    "get",
    "head",
    "options",
    "patch",
    "post",
    "put",
    "request",
    "stream",
]

class _EventHooks(TypedDict):
    request: list[Callable[[httpx.Request], None]]
    response: list[Callable[[httpx.Response], None]]

class _EventHooksParam(TypedDict):
    request: Collection[Callable[[httpx.Request], None]]
    response: Collection[Callable[[httpx.Response], None]]

class _ClientMixin:
    @property
    def auth(self) -> RateLimiter: ...
    @auth.setter
    def auth(self, auth: RateLimiter) -> None: ...
    @property
    def event_hooks(self) -> _EventHooks: ...
    @event_hooks.setter
    def event_hooks(self, value: _EventHooksParam) -> None: ...

class ClientType(_ClientMixin, httpx.Client):
    @overload
    def request(
        self,
        method: str,
        url: URLTypes,
        *,
        content: RequestContent = ...,
        data: RequestData[Any, Any] = ...,
        files: RequestFiles = ...,
        json: Any = ...,
        params: QueryParamTypes = ...,
        headers: HeaderTypes = ...,
        cookies: CookieTypes = ...,
        auth: AuthTypes,
        follow_redirects: bool = ...,
        timeout: TimeoutTypes = ...,
        extensions: dict[Any, Any] = ...,
    ) -> httpx.Response: ...
    @overload
    def request(
        self,
        method: str,
        url: URLTypes,
        *,
        content: RequestContent = ...,
        data: RequestData[Any, Any] = ...,
        files: RequestFiles = ...,
        json: Any = ...,
        params: QueryParamTypes = ...,
        headers: HeaderTypes = ...,
        cookies: CookieTypes = ...,
        auth: RateLimiter = ...,
        follow_redirects: bool = ...,
        timeout: TimeoutTypes = ...,
        extensions: dict[Any, Any] = ...,
    ) -> Response: ...
    @overload
    def send(
        self,
        request: httpx.Request,
        *,
        stream: bool = ...,
        auth: AuthTypes,
        follow_redirects: bool = ...,
    ) -> httpx.Response: ...
    @overload
    def send(
        self,
        request: httpx.Request,
        *,
        stream: bool = ...,
        auth: RateLimiter = ...,
        follow_redirects: bool = ...,
    ) -> Response: ...
    @overload
    def get(
        self,
        url: URLTypes,
        *,
        params: QueryParamTypes = ...,
        headers: HeaderTypes = ...,
        cookies: CookieTypes = ...,
        auth: AuthTypes,
        follow_redirects: bool = ...,
        timeout: TimeoutTypes = ...,
        extensions: dict[Any, Any] = ...,
    ) -> httpx.Response: ...
    @overload
    def get(
        self,
        url: URLTypes,
        *,
        params: QueryParamTypes = ...,
        headers: HeaderTypes = ...,
        cookies: CookieTypes = ...,
        auth: RateLimiter = ...,
        follow_redirects: bool = ...,
        timeout: TimeoutTypes = ...,
        extensions: dict[Any, Any] = ...,
    ) -> Response: ...
    @overload
    def options(
        self,
        url: URLTypes,
        *,
        params: QueryParamTypes = ...,
        headers: HeaderTypes = ...,
        cookies: CookieTypes = ...,
        auth: AuthTypes,
        follow_redirects: bool = ...,
        timeout: TimeoutTypes = ...,
        extensions: dict[Any, Any] = ...,
    ) -> httpx.Response: ...
    @overload
    def options(
        self,
        url: URLTypes,
        *,
        params: QueryParamTypes = ...,
        headers: HeaderTypes = ...,
        cookies: CookieTypes = ...,
        auth: RateLimiter = ...,
        follow_redirects: bool = ...,
        timeout: TimeoutTypes = ...,
        extensions: dict[Any, Any] = ...,
    ) -> Response: ...
    @overload
    def head(
        self,
        url: URLTypes,
        *,
        params: QueryParamTypes = ...,
        headers: HeaderTypes = ...,
        cookies: CookieTypes = ...,
        auth: AuthTypes,
        follow_redirects: bool = ...,
        timeout: TimeoutTypes = ...,
        extensions: dict[Any, Any] = ...,
    ) -> httpx.Response: ...
    @overload
    def head(
        self,
        url: URLTypes,
        *,
        params: QueryParamTypes = ...,
        headers: HeaderTypes = ...,
        cookies: CookieTypes = ...,
        auth: RateLimiter = ...,
        follow_redirects: bool = ...,
        timeout: TimeoutTypes = ...,
        extensions: dict[Any, Any] = ...,
    ) -> Response: ...
    @overload
    def post(
        self,
        url: URLTypes,
        *,
        content: RequestContent = ...,
        data: RequestData[Any, Any] = ...,
        files: RequestFiles = ...,
        json: Any = ...,
        params: QueryParamTypes = ...,
        headers: HeaderTypes = ...,
        cookies: CookieTypes = ...,
        auth: AuthTypes,
        follow_redirects: bool = ...,
        timeout: TimeoutTypes = ...,
        extensions: dict[Any, Any] = ...,
    ) -> httpx.Response: ...
    @overload
    def post(
        self,
        url: URLTypes,
        *,
        content: RequestContent = ...,
        data: RequestData[Any, Any] = ...,
        files: RequestFiles = ...,
        json: Any = ...,
        params: QueryParamTypes = ...,
        headers: HeaderTypes = ...,
        cookies: CookieTypes = ...,
        auth: RateLimiter = ...,
        follow_redirects: bool = ...,
        timeout: TimeoutTypes = ...,
        extensions: dict[Any, Any] = ...,
    ) -> Response: ...
    @overload
    def put(
        self,
        url: URLTypes,
        *,
        content: RequestContent = ...,
        data: RequestData[Any, Any] = ...,
        files: RequestFiles = ...,
        json: Any = ...,
        params: QueryParamTypes = ...,
        headers: HeaderTypes = ...,
        cookies: CookieTypes = ...,
        auth: AuthTypes,
        follow_redirects: bool = ...,
        timeout: TimeoutTypes = ...,
        extensions: dict[Any, Any] = ...,
    ) -> httpx.Response: ...
    @overload
    def put(
        self,
        url: URLTypes,
        *,
        content: RequestContent = ...,
        data: RequestData[Any, Any] = ...,
        files: RequestFiles = ...,
        json: Any = ...,
        params: QueryParamTypes = ...,
        headers: HeaderTypes = ...,
        cookies: CookieTypes = ...,
        auth: RateLimiter = ...,
        follow_redirects: bool = ...,
        timeout: TimeoutTypes = ...,
        extensions: dict[Any, Any] = ...,
    ) -> Response: ...
    @overload
    def patch(
        self,
        url: URLTypes,
        *,
        content: RequestContent = ...,
        data: RequestData[Any, Any] = ...,
        files: RequestFiles = ...,
        json: Any = ...,
        params: QueryParamTypes = ...,
        headers: HeaderTypes = ...,
        cookies: CookieTypes = ...,
        auth: AuthTypes,
        follow_redirects: bool = ...,
        timeout: TimeoutTypes = ...,
        extensions: dict[Any, Any] = ...,
    ) -> httpx.Response: ...
    @overload
    def patch(
        self,
        url: URLTypes,
        *,
        content: RequestContent = ...,
        data: RequestData[Any, Any] = ...,
        files: RequestFiles = ...,
        json: Any = ...,
        params: QueryParamTypes = ...,
        headers: HeaderTypes = ...,
        cookies: CookieTypes = ...,
        auth: RateLimiter = ...,
        follow_redirects: bool = ...,
        timeout: TimeoutTypes = ...,
        extensions: dict[Any, Any] = ...,
    ) -> Response: ...
    @overload
    def delete(
        self,
        url: URLTypes,
        *,
        params: QueryParamTypes = ...,
        headers: HeaderTypes = ...,
        cookies: CookieTypes = ...,
        auth: AuthTypes,
        follow_redirects: bool = ...,
        timeout: TimeoutTypes = ...,
        extensions: dict[Any, Any] = ...,
    ) -> httpx.Response: ...
    @overload
    def delete(
        self,
        url: URLTypes,
        *,
        params: QueryParamTypes = ...,
        headers: HeaderTypes = ...,
        cookies: CookieTypes = ...,
        auth: RateLimiter = ...,
        follow_redirects: bool = ...,
        timeout: TimeoutTypes = ...,
        extensions: dict[Any, Any] = ...,
    ) -> Response: ...
    @overload
    def stream(
        self,
        method: str,
        url: URLTypes,
        *,
        content: RequestContent = ...,
        data: RequestData[Any, Any] = ...,
        files: RequestFiles = ...,
        json: Any = ...,
        params: QueryParamTypes = ...,
        headers: HeaderTypes = ...,
        cookies: CookieTypes = ...,
        auth: AuthTypes,
        follow_redirects: bool = ...,
        timeout: TimeoutTypes = ...,
        extensions: dict[Any, Any] = ...,
    ) -> ContextManager[httpx.Response]: ...
    @overload
    def stream(
        self,
        method: str,
        url: URLTypes,
        *,
        content: RequestContent = ...,
        data: RequestData[Any, Any] = ...,
        files: RequestFiles = ...,
        json: Any = ...,
        params: QueryParamTypes = ...,
        headers: HeaderTypes = ...,
        cookies: CookieTypes = ...,
        auth: RateLimiter = ...,
        follow_redirects: bool = ...,
        timeout: TimeoutTypes = ...,
        extensions: dict[Any, Any] = ...,
    ) -> ContextManager[Response]: ...

class AsyncClientType(_ClientMixin, httpx.AsyncClient):
    @overload
    async def request(
        self,
        method: str,
        url: URLTypes,
        *,
        content: RequestContent = ...,
        data: RequestData[Any, Any] = ...,
        files: RequestFiles = ...,
        json: Any = ...,
        params: QueryParamTypes = ...,
        headers: HeaderTypes = ...,
        cookies: CookieTypes = ...,
        auth: AuthTypes,
        follow_redirects: bool = ...,
        timeout: TimeoutTypes = ...,
        extensions: dict[Any, Any] = ...,
    ) -> httpx.Response: ...
    @overload
    async def request(
        self,
        method: str,
        url: URLTypes,
        *,
        content: RequestContent = ...,
        data: RequestData[Any, Any] = ...,
        files: RequestFiles = ...,
        json: Any = ...,
        params: QueryParamTypes = ...,
        headers: HeaderTypes = ...,
        cookies: CookieTypes = ...,
        auth: RateLimiter = ...,
        follow_redirects: bool = ...,
        timeout: TimeoutTypes = ...,
        extensions: dict[Any, Any] = ...,
    ) -> Response: ...
    @overload
    async def send(
        self,
        request: httpx.Request,
        *,
        stream: bool = ...,
        auth: AuthTypes,
        follow_redirects: bool = ...,
    ) -> httpx.Response: ...
    @overload
    async def send(
        self,
        request: httpx.Request,
        *,
        stream: bool = ...,
        auth: RateLimiter = ...,
        follow_redirects: bool = ...,
    ) -> Response: ...
    @overload
    async def get(
        self,
        url: URLTypes,
        *,
        params: QueryParamTypes = ...,
        headers: HeaderTypes = ...,
        cookies: CookieTypes = ...,
        auth: AuthTypes,
        follow_redirects: bool = ...,
        timeout: TimeoutTypes = ...,
        extensions: dict[Any, Any] = ...,
    ) -> httpx.Response: ...
    @overload
    async def get(
        self,
        url: URLTypes,
        *,
        params: QueryParamTypes = ...,
        headers: HeaderTypes = ...,
        cookies: CookieTypes = ...,
        auth: RateLimiter = ...,
        follow_redirects: bool = ...,
        timeout: TimeoutTypes = ...,
        extensions: dict[Any, Any] = ...,
    ) -> Response: ...
    @overload
    async def options(
        self,
        url: URLTypes,
        *,
        params: QueryParamTypes = ...,
        headers: HeaderTypes = ...,
        cookies: CookieTypes = ...,
        auth: AuthTypes,
        follow_redirects: bool = ...,
        timeout: TimeoutTypes = ...,
        extensions: dict[Any, Any] = ...,
    ) -> httpx.Response: ...
    @overload
    async def options(
        self,
        url: URLTypes,
        *,
        params: QueryParamTypes = ...,
        headers: HeaderTypes = ...,
        cookies: CookieTypes = ...,
        auth: RateLimiter = ...,
        follow_redirects: bool = ...,
        timeout: TimeoutTypes = ...,
        extensions: dict[Any, Any] = ...,
    ) -> Response: ...
    @overload
    async def head(
        self,
        url: URLTypes,
        *,
        params: QueryParamTypes = ...,
        headers: HeaderTypes = ...,
        cookies: CookieTypes = ...,
        auth: AuthTypes,
        follow_redirects: bool = ...,
        timeout: TimeoutTypes = ...,
        extensions: dict[Any, Any] = ...,
    ) -> httpx.Response: ...
    @overload
    async def head(
        self,
        url: URLTypes,
        *,
        params: QueryParamTypes = ...,
        headers: HeaderTypes = ...,
        cookies: CookieTypes = ...,
        auth: RateLimiter = ...,
        follow_redirects: bool = ...,
        timeout: TimeoutTypes = ...,
        extensions: dict[Any, Any] = ...,
    ) -> Response: ...
    @overload
    async def post(
        self,
        url: URLTypes,
        *,
        content: RequestContent = ...,
        data: RequestData[Any, Any] = ...,
        files: RequestFiles = ...,
        json: Any = ...,
        params: QueryParamTypes = ...,
        headers: HeaderTypes = ...,
        cookies: CookieTypes = ...,
        auth: AuthTypes,
        follow_redirects: bool = ...,
        timeout: TimeoutTypes = ...,
        extensions: dict[Any, Any] = ...,
    ) -> httpx.Response: ...
    @overload
    async def post(
        self,
        url: URLTypes,
        *,
        content: RequestContent = ...,
        data: RequestData[Any, Any] = ...,
        files: RequestFiles = ...,
        json: Any = ...,
        params: QueryParamTypes = ...,
        headers: HeaderTypes = ...,
        cookies: CookieTypes = ...,
        auth: RateLimiter = ...,
        follow_redirects: bool = ...,
        timeout: TimeoutTypes = ...,
        extensions: dict[Any, Any] = ...,
    ) -> Response: ...
    @overload
    async def put(
        self,
        url: URLTypes,
        *,
        content: RequestContent = ...,
        data: RequestData[Any, Any] = ...,
        files: RequestFiles = ...,
        json: Any = ...,
        params: QueryParamTypes = ...,
        headers: HeaderTypes = ...,
        cookies: CookieTypes = ...,
        auth: AuthTypes,
        follow_redirects: bool = ...,
        timeout: TimeoutTypes = ...,
        extensions: dict[Any, Any] = ...,
    ) -> httpx.Response: ...
    @overload
    async def put(
        self,
        url: URLTypes,
        *,
        content: RequestContent = ...,
        data: RequestData[Any, Any] = ...,
        files: RequestFiles = ...,
        json: Any = ...,
        params: QueryParamTypes = ...,
        headers: HeaderTypes = ...,
        cookies: CookieTypes = ...,
        auth: RateLimiter = ...,
        follow_redirects: bool = ...,
        timeout: TimeoutTypes = ...,
        extensions: dict[Any, Any] = ...,
    ) -> Response: ...
    @overload
    async def patch(
        self,
        url: URLTypes,
        *,
        content: RequestContent = ...,
        data: RequestData[Any, Any] = ...,
        files: RequestFiles = ...,
        json: Any = ...,
        params: QueryParamTypes = ...,
        headers: HeaderTypes = ...,
        cookies: CookieTypes = ...,
        auth: AuthTypes,
        follow_redirects: bool = ...,
        timeout: TimeoutTypes = ...,
        extensions: dict[Any, Any] = ...,
    ) -> httpx.Response: ...
    @overload
    async def patch(
        self,
        url: URLTypes,
        *,
        content: RequestContent = ...,
        data: RequestData[Any, Any] = ...,
        files: RequestFiles = ...,
        json: Any = ...,
        params: QueryParamTypes = ...,
        headers: HeaderTypes = ...,
        cookies: CookieTypes = ...,
        auth: RateLimiter = ...,
        follow_redirects: bool = ...,
        timeout: TimeoutTypes = ...,
        extensions: dict[Any, Any] = ...,
    ) -> Response: ...
    @overload
    async def delete(
        self,
        url: URLTypes,
        *,
        params: QueryParamTypes = ...,
        headers: HeaderTypes = ...,
        cookies: CookieTypes = ...,
        auth: AuthTypes,
        follow_redirects: bool = ...,
        timeout: TimeoutTypes = ...,
        extensions: dict[Any, Any] = ...,
    ) -> httpx.Response: ...
    @overload
    async def delete(
        self,
        url: URLTypes,
        *,
        params: QueryParamTypes = ...,
        headers: HeaderTypes = ...,
        cookies: CookieTypes = ...,
        auth: RateLimiter = ...,
        follow_redirects: bool = ...,
        timeout: TimeoutTypes = ...,
        extensions: dict[Any, Any] = ...,
    ) -> Response: ...
    @overload
    def stream(
        self,
        method: str,
        url: URLTypes,
        *,
        content: RequestContent = ...,
        data: RequestData[Any, Any] = ...,
        files: RequestFiles = ...,
        json: Any = ...,
        params: QueryParamTypes = ...,
        headers: HeaderTypes = ...,
        cookies: CookieTypes = ...,
        auth: AuthTypes,
        follow_redirects: bool = ...,
        timeout: TimeoutTypes = ...,
        extensions: dict[Any, Any] = ...,
    ) -> AsyncContextManager[httpx.Response]: ...
    @overload
    def stream(
        self,
        method: str,
        url: URLTypes,
        *,
        content: RequestContent = ...,
        data: RequestData[Any, Any] = ...,
        files: RequestFiles = ...,
        json: Any = ...,
        params: QueryParamTypes = ...,
        headers: HeaderTypes = ...,
        cookies: CookieTypes = ...,
        auth: RateLimiter = ...,
        follow_redirects: bool = ...,
        timeout: TimeoutTypes = ...,
        extensions: dict[Any, Any] = ...,
    ) -> AsyncContextManager[Response]: ...

@overload
def Client(
    *,
    auth: AuthTypes,
    params: QueryParamTypes = ...,
    headers: HeaderTypes = ...,
    cookies: CookieTypes = ...,
    verify: VerifyTypes = ...,
    cert: CertTypes = ...,
    http1: bool = ...,
    http2: bool = ...,
    proxies: ProxiesTypes = ...,
    mounts: Mapping[str, httpx.BaseTransport] = ...,
    timeout: TimeoutTypes = ...,
    follow_redirects: bool = ...,
    limits: httpx.Limits = ...,
    max_redirects: int = ...,
    event_hooks: Mapping[str, list[Callable[..., Any]]] = ...,
    base_url: URLTypes = ...,
    transport: httpx.BaseTransport = ...,
    app: Callable[..., Any] = ...,
    trust_env: bool = ...,
) -> httpx.Client: ...
@overload
def Client(
    *,
    auth: RateLimiter = ...,
    params: QueryParamTypes = ...,
    headers: HeaderTypes = ...,
    cookies: CookieTypes = ...,
    verify: VerifyTypes = ...,
    cert: CertTypes = ...,
    http1: bool = ...,
    http2: bool = ...,
    proxies: ProxiesTypes = ...,
    mounts: Mapping[str, httpx.BaseTransport] = ...,
    timeout: TimeoutTypes = ...,
    follow_redirects: bool = ...,
    limits: httpx.Limits = ...,
    max_redirects: int = ...,
    event_hooks: _EventHooksParam = ...,
    base_url: URLTypes = ...,
    transport: httpx.BaseTransport = ...,
    app: Callable[..., Any] = ...,
    trust_env: bool = ...,
) -> ClientType: ...
@overload
def AsyncClient(
    *,
    auth: AuthTypes,
    params: QueryParamTypes = ...,
    headers: HeaderTypes = ...,
    cookies: CookieTypes = ...,
    verify: VerifyTypes = ...,
    cert: CertTypes = ...,
    http1: bool = ...,
    http2: bool = ...,
    proxies: ProxiesTypes = ...,
    mounts: Mapping[str, httpx.BaseTransport] = ...,
    timeout: TimeoutTypes = ...,
    follow_redirects: bool = ...,
    limits: httpx.Limits = ...,
    max_redirects: int = ...,
    event_hooks: _EventHooksParam = ...,
    base_url: URLTypes = ...,
    transport: httpx.BaseTransport = ...,
    app: Callable[..., Any] = ...,
    trust_env: bool = ...,
) -> httpx.AsyncClient: ...
@overload
def AsyncClient(
    *,
    auth: RateLimiter = ...,
    params: QueryParamTypes = ...,
    headers: HeaderTypes = ...,
    cookies: CookieTypes = ...,
    verify: VerifyTypes = ...,
    cert: CertTypes = ...,
    http1: bool = ...,
    http2: bool = ...,
    proxies: ProxiesTypes = ...,
    mounts: Mapping[str, httpx.BaseTransport] = ...,
    timeout: TimeoutTypes = ...,
    follow_redirects: bool = ...,
    limits: httpx.Limits = ...,
    max_redirects: int = ...,
    event_hooks: Mapping[str, Collection[Callable[..., Any]]] = ...,
    base_url: URLTypes = ...,
    transport: httpx.BaseTransport = ...,
    app: Callable[..., Any] = ...,
    trust_env: bool = ...,
) -> AsyncClientType: ...
@overload
def request(
    method: str,
    url: URLTypes,
    *,
    params: QueryParamTypes = ...,
    content: RequestContent = ...,
    data: RequestData[Any, Any] = ...,
    files: RequestFiles = ...,
    json: Any = ...,
    headers: HeaderTypes = ...,
    cookies: CookieTypes = ...,
    auth: AuthTypes,
    proxies: ProxiesTypes = ...,
    timeout: TimeoutTypes = ...,
    follow_redirects: bool = ...,
    verify: VerifyTypes = ...,
    cert: CertTypes = ...,
    trust_env: bool = ...,
) -> httpx.Response: ...
@overload
def request(
    method: str,
    url: URLTypes,
    *,
    params: QueryParamTypes = ...,
    content: RequestContent = ...,
    data: RequestData[Any, Any] = ...,
    files: RequestFiles = ...,
    json: Any = ...,
    headers: HeaderTypes = ...,
    cookies: CookieTypes = ...,
    auth: RateLimiter = ...,
    proxies: ProxiesTypes = ...,
    timeout: TimeoutTypes = ...,
    follow_redirects: bool = ...,
    verify: VerifyTypes = ...,
    cert: CertTypes = ...,
    trust_env: bool = ...,
) -> Response: ...
@overload
def stream(
    method: str,
    url: URLTypes,
    *,
    params: QueryParamTypes = ...,
    content: RequestContent = ...,
    data: RequestData[Any, Any] = ...,
    files: RequestFiles = ...,
    json: Any = ...,
    headers: HeaderTypes = ...,
    cookies: CookieTypes = ...,
    auth: AuthTypes,
    proxies: ProxiesTypes = ...,
    timeout: TimeoutTypes = ...,
    follow_redirects: bool = ...,
    verify: VerifyTypes = ...,
    cert: CertTypes = ...,
    trust_env: bool = ...,
) -> ContextManager[httpx.Response]: ...
@overload
def stream(
    method: str,
    url: URLTypes,
    *,
    params: QueryParamTypes = ...,
    content: RequestContent = ...,
    data: RequestData[Any, Any] = ...,
    files: RequestFiles = ...,
    json: Any = ...,
    headers: HeaderTypes = ...,
    cookies: CookieTypes = ...,
    auth: RateLimiter = ...,
    proxies: ProxiesTypes = ...,
    timeout: TimeoutTypes = ...,
    follow_redirects: bool = ...,
    verify: VerifyTypes = ...,
    cert: CertTypes = ...,
    trust_env: bool = ...,
) -> ContextManager[Response]: ...
@overload
def get(
    url: URLTypes,
    *,
    params: QueryParamTypes = ...,
    headers: HeaderTypes = ...,
    cookies: CookieTypes = ...,
    auth: AuthTypes,
    proxies: ProxiesTypes = ...,
    follow_redirects: bool = ...,
    cert: CertTypes = ...,
    verify: VerifyTypes = ...,
    timeout: TimeoutTypes = ...,
    trust_env: bool = ...,
) -> httpx.Response: ...
@overload
def get(
    url: URLTypes,
    *,
    params: QueryParamTypes = ...,
    headers: HeaderTypes = ...,
    cookies: CookieTypes = ...,
    auth: RateLimiter = ...,
    proxies: ProxiesTypes = ...,
    follow_redirects: bool = ...,
    cert: CertTypes = ...,
    verify: VerifyTypes = ...,
    timeout: TimeoutTypes = ...,
    trust_env: bool = ...,
) -> Response: ...
@overload
def options(
    url: URLTypes,
    *,
    params: QueryParamTypes = ...,
    headers: HeaderTypes = ...,
    cookies: CookieTypes = ...,
    auth: AuthTypes,
    proxies: ProxiesTypes = ...,
    follow_redirects: bool = ...,
    cert: CertTypes = ...,
    verify: VerifyTypes = ...,
    timeout: TimeoutTypes = ...,
    trust_env: bool = ...,
) -> httpx.Response: ...
@overload
def options(
    url: URLTypes,
    *,
    params: QueryParamTypes = ...,
    headers: HeaderTypes = ...,
    cookies: CookieTypes = ...,
    auth: RateLimiter = ...,
    proxies: ProxiesTypes = ...,
    follow_redirects: bool = ...,
    cert: CertTypes = ...,
    verify: VerifyTypes = ...,
    timeout: TimeoutTypes = ...,
    trust_env: bool = ...,
) -> Response: ...
@overload
def head(
    url: URLTypes,
    *,
    params: QueryParamTypes = ...,
    headers: HeaderTypes = ...,
    cookies: CookieTypes = ...,
    auth: AuthTypes,
    proxies: ProxiesTypes = ...,
    follow_redirects: bool = ...,
    cert: CertTypes = ...,
    verify: VerifyTypes = ...,
    timeout: TimeoutTypes = ...,
    trust_env: bool = ...,
) -> httpx.Response: ...
@overload
def head(
    url: URLTypes,
    *,
    params: QueryParamTypes = ...,
    headers: HeaderTypes = ...,
    cookies: CookieTypes = ...,
    auth: RateLimiter = ...,
    proxies: ProxiesTypes = ...,
    follow_redirects: bool = ...,
    cert: CertTypes = ...,
    verify: VerifyTypes = ...,
    timeout: TimeoutTypes = ...,
    trust_env: bool = ...,
) -> Response: ...
@overload
def post(
    url: URLTypes,
    *,
    params: QueryParamTypes = ...,
    content: RequestContent = ...,
    data: RequestData[Any, Any] = ...,
    files: RequestFiles = ...,
    json: Any = ...,
    headers: HeaderTypes = ...,
    cookies: CookieTypes = ...,
    auth: AuthTypes,
    proxies: ProxiesTypes = ...,
    timeout: TimeoutTypes = ...,
    follow_redirects: bool = ...,
    verify: VerifyTypes = ...,
    cert: CertTypes = ...,
    trust_env: bool = ...,
) -> httpx.Response: ...
@overload
def post(
    url: URLTypes,
    *,
    params: QueryParamTypes = ...,
    content: RequestContent = ...,
    data: RequestData[Any, Any] = ...,
    files: RequestFiles = ...,
    json: Any = ...,
    headers: HeaderTypes = ...,
    cookies: CookieTypes = ...,
    auth: RateLimiter = ...,
    proxies: ProxiesTypes = ...,
    timeout: TimeoutTypes = ...,
    follow_redirects: bool = ...,
    verify: VerifyTypes = ...,
    cert: CertTypes = ...,
    trust_env: bool = ...,
) -> Response: ...
@overload
def put(
    url: URLTypes,
    *,
    params: QueryParamTypes = ...,
    content: RequestContent = ...,
    data: RequestData[Any, Any] = ...,
    files: RequestFiles = ...,
    json: Any = ...,
    headers: HeaderTypes = ...,
    cookies: CookieTypes = ...,
    auth: AuthTypes,
    proxies: ProxiesTypes = ...,
    timeout: TimeoutTypes = ...,
    follow_redirects: bool = ...,
    verify: VerifyTypes = ...,
    cert: CertTypes = ...,
    trust_env: bool = ...,
) -> httpx.Response: ...
@overload
def put(
    url: URLTypes,
    *,
    params: QueryParamTypes = ...,
    content: RequestContent = ...,
    data: RequestData[Any, Any] = ...,
    files: RequestFiles = ...,
    json: Any = ...,
    headers: HeaderTypes = ...,
    cookies: CookieTypes = ...,
    auth: RateLimiter = ...,
    proxies: ProxiesTypes = ...,
    timeout: TimeoutTypes = ...,
    follow_redirects: bool = ...,
    verify: VerifyTypes = ...,
    cert: CertTypes = ...,
    trust_env: bool = ...,
) -> Response: ...
@overload
def patch(
    url: URLTypes,
    *,
    params: QueryParamTypes = ...,
    content: RequestContent = ...,
    data: RequestData[Any, Any] = ...,
    files: RequestFiles = ...,
    json: Any = ...,
    headers: HeaderTypes = ...,
    cookies: CookieTypes = ...,
    auth: AuthTypes,
    proxies: ProxiesTypes = ...,
    timeout: TimeoutTypes = ...,
    follow_redirects: bool = ...,
    verify: VerifyTypes = ...,
    cert: CertTypes = ...,
    trust_env: bool = ...,
) -> httpx.Response: ...
@overload
def patch(
    url: URLTypes,
    *,
    params: QueryParamTypes = ...,
    content: RequestContent = ...,
    data: RequestData[Any, Any] = ...,
    files: RequestFiles = ...,
    json: Any = ...,
    headers: HeaderTypes = ...,
    cookies: CookieTypes = ...,
    auth: RateLimiter = ...,
    proxies: ProxiesTypes = ...,
    timeout: TimeoutTypes = ...,
    follow_redirects: bool = ...,
    verify: VerifyTypes = ...,
    cert: CertTypes = ...,
    trust_env: bool = ...,
) -> Response: ...
@overload
def delete(
    url: URLTypes,
    *,
    params: QueryParamTypes = ...,
    headers: HeaderTypes = ...,
    cookies: CookieTypes = ...,
    auth: AuthTypes,
    proxies: ProxiesTypes = ...,
    follow_redirects: bool = ...,
    cert: CertTypes = ...,
    verify: VerifyTypes = ...,
    timeout: TimeoutTypes = ...,
    trust_env: bool = ...,
) -> httpx.Response: ...
@overload
def delete(
    url: URLTypes,
    *,
    params: QueryParamTypes = ...,
    headers: HeaderTypes = ...,
    cookies: CookieTypes = ...,
    auth: RateLimiter = ...,
    proxies: ProxiesTypes = ...,
    follow_redirects: bool = ...,
    cert: CertTypes = ...,
    verify: VerifyTypes = ...,
    timeout: TimeoutTypes = ...,
    trust_env: bool = ...,
) -> Response: ...



from __future__ import annotations

import codecs
import zlib
from operator import itemgetter
from typing import TYPE_CHECKING, Iterable
from xml.etree.ElementTree import Element, XMLParser, XMLPullParser

if TYPE_CHECKING:
    from typing_extensions import Literal


class GZipDecoder:
    def __init__(self) -> None:
        self._decompressobj = zlib.decompressobj(zlib.MAX_WBITS | 16)

    def decode(self, data: bytes) -> bytes:
        return self._decompressobj.decompress(data)

    def flush(self) -> bytes:
        return self._decompressobj.flush()


class XMLDecoder:
    def __init__(self, encoding: str | None = None) -> None:
        self._parser = XMLParser(encoding=encoding)

    def decode(self, data: bytes) -> None:
        self._parser.feed(data)

    def flush(self) -> Element:
        return self._parser.close()


class XMLChunker:
    def __init__(self, *, encoding: str | None = None) -> None:
        self._decoder_chain: list[codecs.IncrementalDecoder] = []
        if encoding:
            self._decoder_chain.append(
                codecs.getincrementaldecoder(encoding)("replace")
            )
        self._pull_parser = XMLPullParser(["start", "end"])
        self._root: Element | None = None

    def decode(self, data: bytes) -> Iterable[Element]:
        for chain in self._decoder_chain:
            data = chain.decode(data, False)  # type: ignore
        self._pull_parser.feed(data)
        return map(itemgetter(1), self._read_events())

    def flush(self, data: bytes = b"") -> Iterable[Element]:
        for chain in self._decoder_chain:
            data = chain.decode(data, True)  # type: ignore
        self._pull_parser.feed(data)
        self._pull_parser.close()
        return map(itemgetter(1), self._read_events())

    def _read_events(self) -> Iterable[tuple[Literal["end"], Element]]:
        depth = 0
        element: Element
        for event, element in self._pull_parser.read_events():
            if not self._root:
                self._root = element
            if event == "start":
                depth += 1
            elif event == "end":
                depth -= 1
                if depth == 1:
                    yield event, element
                    self._root.clear()


try:
    from lxml.etree import (
        XMLParser as LXMLParser,
        XMLPullParser as LXMLPullParser,
        _Element as LElement,  # type: ignore
    )
except ImportError:
    pass
else:

    class LXMLDecoder:
        def __init__(self, encoding: str | None = None) -> None:
            self._parser = LXMLParser(encoding=encoding)

        def decode(self, data: bytes) -> None:
            self._parser.feed(data)

        def flush(self) -> LElement:
            return self._parser.close()

    class LXMLChunker:
        def __init__(self, *, encoding: str | None = None) -> None:
            self._decoder_chain: list[codecs.IncrementalDecoder] = []
            if encoding:
                self._decoder_chain.append(
                    codecs.getincrementaldecoder(encoding)("replace")
                )
            self._pull_parser = LXMLPullParser(["start", "end"])

        def decode(self, data: bytes) -> Iterable[LElement]:
            for chain in self._decoder_chain:
                data = chain.decode(data, False)  # type: ignore
            self._pull_parser.feed(data)
            return map(itemgetter(1), self._read_events())

        def flush(self, data: bytes = b"") -> Iterable[LElement]:
            for chain in self._decoder_chain:
                data = chain.decode(data, True)  # type: ignore
            self._pull_parser.feed(data)
            self._pull_parser.close()
            return map(itemgetter(1), self._read_events())

        def _read_events(self) -> Iterable[tuple[Literal["end"], LElement]]:
            depth = 0
            element: LElement
            for event, element in self._pull_parser.read_events():  # type: ignore
                if not self._root:
                    self._root = element
                if event == "start":
                    depth += 1
                elif event == "end":
                    depth -= 1
                    if depth == 1:
                        yield event, element
                        self._root.clear()



from __future__ import annotations

from httpx import Request, Response

from .limiter import RateLimiter

try:
    # guard against reloading
    _PINS  # type: ignore  # noqa: B018
except NameError:
    _PINS: dict[str, str] = {}


__all__ = ["NSAuth"]


class NSAuth(RateLimiter):
    def __init__(self, *, password: str | None = None, autologin: str | None = None):
        self._password = password
        self.autologin = autologin

    @property
    def autologin(self) -> str | None:
        return self._autologin

    @autologin.setter
    def autologin(self, value: str | None) -> None:
        if value:
            self._password = None  # no longer needed
        self._autologin = value

    def _request_hook(self, request: Request) -> Request:
        nation = request.url.params.get("nation")
        if not nation:
            return request
        auth = {}
        if self._password:
            auth["X-Password"] = self._password
        elif self._autologin:
            auth["X-Autologin"] = self._autologin
        pin = _PINS.get(nation)
        if pin:
            auth["X-Pin"] = pin
        request.method = "POST"
        return super()._request_hook(request)

    def _response_hook(self, response: Response) -> None:
        autologin = response.headers.get("X-Autologin")
        if autologin:
            self.autologin = autologin
        return super()._response_hook(response)



# MIT License

# Copyright (c) 2018 - 2023

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import sys as _sys

from ._state import set_agent as set_agent
from .auth import *
from .client import *
from .errors import *
from .limiter import *
from .lock import *
from .response import *
from .url import *
from .utils import *

if _sys.version_info < (3, 8):
    from importlib_metadata import metadata as _metadata  # type: ignore
else:
    from importlib.metadata import metadata as _metadata

__copyright__ = "Copyright 2019-2023 Zephyrkul"

_meta = _metadata(__name__)
__title__ = _meta["Name"]
__author__ = _meta["Author"]
__license__ = _meta["License"]
__version__ = _meta["Version"]

del _sys, _metadata, _meta  # not for export



from __future__ import annotations

from datetime import date as _date
from typing import TYPE_CHECKING, Mapping

import httpx

if TYPE_CHECKING:
    from typing_extensions import Literal
API_URL = httpx.URL("https://www.nationstates.net/cgi-bin/api.cgi")

__all__ = [
    "API_URL",
    "Nation",
    "Region",
    "World",
    "WA",
    "Shard",
    "NationsDump",
    "RegionsDump",
    "CardsDump",
]


def Nation(nation: str, *shards: str, **parameters: str) -> httpx.URL:
    return World(*shards, nation=nation, **parameters)


def Region(region: str, *shards: str, **parameters: str) -> httpx.URL:
    return World(*shards, region=region, **parameters)


def World(*shards: str | Mapping[str, str], **parameters: str) -> httpx.URL:
    q: list[str | None] = [parameters.pop("q", None)]
    query: dict[str, str] = {}
    for shard in shards:
        try:
            shard = dict(shard)  # type: ignore
        except (TypeError, ValueError):
            q.append(str(shard))
        else:
            q.append(shard.pop("q", None))  # type: ignore
            query.update(shard)  # type: ignore
    query.update(parameters, q="+".join(map(str, filter(None, q))))
    if query.get("a", "").lower() == "sendtg":
        raise RuntimeError("sans does not currently support the telegram API.")
    return API_URL.copy_with(params=query)


def WA(wa: Literal[1, "1", 2, "2"], *shards: str, **parameters: str) -> httpx.URL:
    return World(*shards, wa=str(wa), **parameters)


def Shard(q: str, **parameters: str) -> dict[str, str]:
    parameters["q"] = q
    return parameters


# https://www.nationstates.net/archive/nations/2018-09-30-nations-xml.gz
def NationsDump(date: _date | None = None) -> httpx.URL:
    if date:
        path = date.strftime("/archive/nations/%Y-%m-%d-nations-xml.gz")
    else:
        path = "/pages/nations.xml.gz"
    return API_URL.join(path)


# https://www.nationstates.net/archive/nations/2018-09-30-regions-xml.gz
def RegionsDump(date: _date | None = None) -> httpx.URL:
    if date:
        path = date.strftime("/archive/nations/%Y-%m-%d-regions-xml.gz")
    else:
        path = "/pages/regions.xml.gz"
    return API_URL.join(path)


def CardsDump(season: Literal[1, "1", 2, "2", 3, "3"]) -> httpx.URL:
    return API_URL.join(f"/pages/cardlist_S{season}.xml.gz")



from __future__ import annotations

from functools import wraps
from typing import TYPE_CHECKING, Callable, TypeVar

if TYPE_CHECKING:
    from typing_extensions import ParamSpec

    _P = ParamSpec("_P")
    _R = TypeVar("_R")

import httpx
from httpx import AsyncClient as AsyncClientType, Client as ClientType

from .limiter import RateLimiter

__all__ = [
    "Client",
    "AsyncClient",
    "delete",
    "get",
    "head",
    "options",
    "patch",
    "post",
    "put",
    "request",
    "stream",
]
_limiter = RateLimiter()


def _ensure_auth(wrapped: Callable[_P, _R]) -> Callable[_P, _R]:
    @wraps(wrapped)
    def inner(*args: _P.args, **kwargs: _P.kwargs) -> _R:
        if "auth" not in kwargs:
            kwargs["auth"] = _limiter
        return wrapped(*args, **kwargs)

    return inner


Client = _ensure_auth(ClientType)  # type: ignore
AsyncClient = _ensure_auth(AsyncClientType)  # type: ignore

request = _ensure_auth(httpx.request)  # type: ignore
delete = _ensure_auth(httpx.delete)
get = _ensure_auth(httpx.get)
head = _ensure_auth(httpx.head)
options = _ensure_auth(httpx.options)
patch = _ensure_auth(httpx.patch)  # type: ignore
post = _ensure_auth(httpx.post)  # type: ignore
put = _ensure_auth(httpx.put)  # type: ignore
stream = _ensure_auth(httpx.stream)  # type: ignore



from __future__ import annotations

from email.message import Message
from itertools import chain
from typing import Any, AsyncIterator, Iterator
from xml.etree.ElementTree import Element

import httpx

from .decoder import GZipDecoder, XMLChunker, XMLDecoder
from .errors import narrow

try:
    import xmltodict

    HAS_XMLTODICT = True
except ImportError:
    HAS_XMLTODICT = False  # type: ignore

try:
    from lxml.etree import _Element  # type: ignore

    HAS_LXML = True
except ImportError:
    HAS_LXML = False  # type: ignore
else:
    from .decoder import LXMLChunker, LXMLDecoder

__all__ = ["Response"]


class Response(httpx.Response):
    @property
    def content_type(self) -> str:
        if not hasattr(self, "_content_type"):
            message = Message()
            content_type = self.headers.get("Content-Type")
            if content_type:
                message["Content-Type"] = self.headers.get("Content-Type")
            self._content_type = message.get_content_type()
        return self._content_type

    def iter_gzip(self) -> Iterator[bytes]:
        decoder = GZipDecoder()
        yield from map(decoder.decode, self.iter_bytes())
        yield decoder.flush()

    async def aiter_gzip(self) -> AsyncIterator[bytes]:
        decoder = GZipDecoder()
        async for chunk in self.aiter_bytes():
            yield decoder.decode(chunk)
        yield decoder.flush()

    if HAS_XMLTODICT:

        def json(self, **kwargs: Any) -> dict[str, Any]:
            if self.content_type.endswith("/xml"):
                return xmltodict.parse(self.content, encoding=self.encoding, **kwargs)
            return super().json(**kwargs)

    @property
    def xml(self) -> Element:
        if not hasattr(self, "_xml"):
            content = self.content
            decoder = XMLDecoder(self.encoding)
            decoder.decode(content)
            self._xml = decoder.flush()
        return self._xml

    def iter_xml(self) -> Iterator[Element]:
        decoder = XMLChunker(encoding=self.encoding)
        chunker = (
            self.iter_gzip()
            if self.content_type.endswith(("/x-gzip", "/gzip"))
            else self.iter_bytes()
        )
        yield from chain.from_iterable(map(decoder.decode, chunker))
        yield from decoder.flush()

    async def aiter_xml(self) -> AsyncIterator[Element]:
        decoder = XMLChunker(encoding=self.encoding)
        chunker = (
            self.aiter_gzip()
            if self.content_type.endswith(("/x-gzip", "/gzip"))
            else self.aiter_bytes()
        )
        async for chunk in chunker:
            for element in decoder.decode(chunk):
                yield element
        for element in decoder.flush():
            yield element

    if HAS_LXML:

        @property
        def lxml(self) -> _Element:
            if not hasattr(self, "_lxml"):
                content = self.content
                decoder = LXMLDecoder(self.encoding)
                decoder.decode(content)
                self._lxml = decoder.flush()
            return self._lxml

        def iter_lxml(self) -> Iterator[_Element]:
            decoder = LXMLChunker(encoding=self.encoding)
            chunker = (
                self.iter_gzip()
                if self.content_type.endswith(("/x-gzip", "/gzip"))
                else self.iter_bytes()
            )
            yield from chain.from_iterable(map(decoder.decode, chunker))
            yield from decoder.flush()

        async def aiter_lxml(self) -> AsyncIterator[_Element]:
            decoder = LXMLChunker(encoding=self.encoding)
            chunker = (
                self.aiter_gzip()
                if self.content_type.endswith(("/x-gzip", "/gzip"))
                else self.aiter_bytes()
            )
            async for chunk in chunker:
                for element in decoder.decode(chunk):
                    yield element
            for element in decoder.flush():
                yield element

    def raise_for_status(self) -> None:
        try:
            return super().raise_for_status()
        except httpx.HTTPStatusError as exc:
            raise narrow(exc) from None



#!/usr/bin/python3

from __future__ import annotations

import argparse
import shlex
import sys
import traceback
from typing import TYPE_CHECKING, Any
from xml.etree import ElementTree as ET

import sans

try:
    if not TYPE_CHECKING:
        from rich import print  # type: ignore
        from rich.syntax import Syntax  # type: ignore
except ImportError:

    def Syntax(arg: Any, *_args: Any, **_kwargs: Any) -> Any:
        return arg


def main() -> None:
    parser = argparse.ArgumentParser(
        allow_abbrev=False,
        description="SANS Console Entry",
        prog=sans.__name__,
        epilog="Any unknown args will be used to build the API request.",
    )
    parser.add_argument("--agent", help="set the script's user agent")
    parser.add_argument(
        "--headers",
        action="store_true",
        help="print the headers of the server response",
    )
    parser.add_argument(
        "--quit",
        action="store_true",
        help="quit the console program loop after this run",
    )
    parser.add_argument(
        "--url", action="store_true", help="print the URL of the generated request"
    )
    parser.add_argument("--version", action="version", version=sans.__version__)
    input_: list[str] | None = None
    with sans.Client() as client:
        while True:
            print(file=sys.stderr)
            known, unknown = parser.parse_known_args(input_)
            try:
                if known.agent:
                    try:
                        sans.set_agent(known.agent)
                    except RuntimeError:
                        print(
                            "You can't change the agent in the middle of the script.",
                            file=sys.stderr,
                        )
                if not unknown:
                    if known.quit:
                        print("No query provided. Exiting...", file=sys.stderr)
                        return
                    print("No query provided.", file=sys.stderr)
                    parser.print_help(sys.stderr)
                    continue
                parameters: dict[str, list[str]] = {}
                key = "q"
                for arg in unknown:
                    if arg.startswith("--"):
                        if key != "q":
                            print(
                                "No value provided for key {!r}".format(key),
                                file=sys.stderr,
                            )
                        key = arg[2:]
                    else:
                        parameters.setdefault(key, []).append(arg)
                        key = "q"
                if key != "q":
                    print("No value provided for key {!r}".format(key), file=sys.stderr)
                with client.stream(
                    "GET",
                    sans.World(**{k: "+".join(v) for k, v in parameters.items()}),
                ) as response:
                    if known.url:
                        print(response.url, end="\n\n")
                    if known.headers:
                        print(response.headers, end="\n\n")
                    for element in response.iter_xml():
                        pretty_print(element)
                if known.quit:
                    print("Exiting...", file=sys.stderr)
                    return
            except Exception:
                traceback.print_exc()
            except BaseException:
                known.quit = True  # bypass the finally block below
                raise
            finally:
                if not known.quit:
                    input_ = shlex.split(input(f">>> {parser.prog} "))


def pretty_print(element: ET.Element) -> None:
    sans.indent(element)
    print(Syntax(ET.tostring(element, encoding="unicode").strip(), "xml"))


if __name__ == "__main__":
    sys.exit(main())



from __future__ import annotations

import logging
import sys
import time
from itertools import count
from math import ldexp
from random import random
from typing import AsyncGenerator, Generator, Iterator

import anyio
import httpx

from . import _state  # type: ignore
from .errors import AgentNotSetError
from .response import Response
from .url import API_URL

__all__ = ["RateLimiter"]
LOG = logging.getLogger(__name__)


class _Backoff:
    def __getitem__(self, item: slice) -> Iterator[float]:
        def _next(index: int) -> float:
            # Random number in the range of [0, 2**index)
            return ldexp(random(), index)

        if item.stop is None or item.step == 0:
            return map(_next, count(item.start, item.step))

        return map(_next, range(item.start or 0, item.stop, item.step or 1))


class _NullContext:
    def __enter__(self):
        pass

    def __exit__(self, *_):
        pass

    async def __aenter__(self):
        pass

    async def __aexit__(self, *_):
        pass


_backoff = _Backoff()
_nullctx = _NullContext()


class RateLimiter(httpx.Auth):
    """
    httpx Auth utility object which implements ratelimiting, User-Agent management,
    and exponential backoff, while adding XML support to the resulting Response.
    """

    def _request_hook(self, request: httpx.Request) -> httpx.Request:
        agent = _state.agent
        if not agent:
            raise AgentNotSetError("sans has no set agent")
        request.headers["User-Agent"] = agent
        return request

    def _response_hook(self, response: httpx.Response) -> None:
        _state.lock.maybe_defer(response.headers)

    def auth_flow(
        self, request: httpx.Request
    ) -> Generator[httpx.Request, httpx.Response, None]:
        if request.url.host == API_URL.host:
            request = self._request_hook(request)
        response = yield request
        response.__class__ = Response
        if response.url.host == API_URL.host:
            self._response_hook(response)

    def sync_auth_flow(
        self, request: httpx.Request
    ) -> Generator[httpx.Request, httpx.Response, None]:
        if request.url.copy_with(query=None) == API_URL:
            ctx = _state.lock
        else:
            ctx = _nullctx
        with ctx:
            # The below is equivalent to: return (yield from self.auth_flow(request))
            # but with exponential backoff, locking, and ratelimit retry added.
            # See https://peps.python.org/pep-0380/#formal-semantics for details
            flow = self.auth_flow(request)
            try:
                _yield = next(flow)
            except StopIteration as exc:
                return exc.value
            backoff = _backoff[:5]
            while True:
                try:
                    _sent = yield _yield
                except GeneratorExit:
                    flow.close()
                    raise
                except BaseException:
                    try:
                        flow.throw(*sys.exc_info())  # type: ignore
                    except StopIteration as exc:
                        return exc.value
                else:
                    status = _sent.status_code
                    if status == 429:
                        retry = int(_sent.headers.get("Retry-After", 0))
                        if retry:
                            LOG.info("Ratelimit hit, retrying in %s seconds.", retry)
                            time.sleep(retry)
                            continue
                    elif status in [500, 502]:
                        try:
                            retry = next(backoff)
                        except StopIteration:
                            pass  # out of retries
                        else:
                            LOG.debug(
                                "Status %s received, retrying in %.1f seconds.",
                                status,
                                retry,
                            )
                            time.sleep(retry)
                            continue  # retry with the same request
                    try:
                        _yield = flow.send(_sent)
                    except StopIteration as exc:
                        return exc.value
                    backoff = _backoff[:5]

    async def async_auth_flow(
        self, request: httpx.Request
    ) -> AsyncGenerator[httpx.Request, httpx.Response]:
        if request.url.copy_with(query=None) == API_URL:
            ctx = _state.lock
        else:
            ctx = _nullctx
        async with ctx:
            # The below is equivalent to: return (yield from self.auth_flow(request))
            # but with exponential backoff, locking, and ratelimit retry added.
            # See https://peps.python.org/pep-0380/#formal-semantics for details
            flow = self.auth_flow(request)
            try:
                _yield = next(flow)
            except StopIteration as exc:
                assert exc.value is None
                return
            backoff = _backoff[:5]
            while True:
                try:
                    _sent = yield _yield
                except GeneratorExit:
                    flow.close()
                    raise
                except BaseException:
                    try:
                        flow.throw(*sys.exc_info())  # type: ignore
                    except StopIteration as exc:
                        assert exc.value is None
                        return
                else:
                    status = _sent.status_code
                    if status == 429:
                        retry = int(_sent.headers.get("Retry-After", 0))
                        if retry:
                            await anyio.sleep(retry)
                            continue
                    elif status in [500, 502]:
                        try:
                            retry = next(backoff)
                        except StopIteration:
                            pass  # out of retries
                        else:
                            LOG.debug(
                                "Status %s received, retrying in %.0f seconds.",
                                status,
                                retry,
                            )
                            await anyio.sleep(retry)
                            continue  # retry with the same request
                    try:
                        _yield = flow.send(_sent)
                    except StopIteration as exc:
                        assert exc.value is None
                        return
                    backoff = _backoff[:5]
